# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2016-2023 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-instance-attributes

"""Module for flat-review commands"""

from __future__ import annotations

import time
from enum import Enum
from typing import TYPE_CHECKING

import gi

gi.require_version("Atspi", "2.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Atspi, GLib, Gtk  # pylint: disable=no-name-in-module

from . import (
    braille,
    braille_presenter,
    clipboard,
    dbus_service,
    debug,
    flat_review,
    flat_review_presenter_command_definitions,
    focus_manager,
    gsettings_registry,
    guilabels,
    input_event,
    input_event_manager,
    messages,
    presentation_manager,
    script_manager,
    speech_presenter,
)
from .ax_event_synthesizer import AXEventSynthesizer
from .ax_object import AXObject
from .ax_text import AXText
from .ax_utilities import AXUtilities
from .ax_utilities_event import TextEventReason
from .ax_utilities_text import CaretSetReason
from .extension import Extension

if TYPE_CHECKING:
    from .command import Command
    from .dbus_service import UInt32
    from .flat_review_presenter_preferences_grid import FlatReviewPreferencesGrid
    from .scripts import default


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.FocusTracking",
    values={"off": 0, "auto": 1, "on": 2},
)
class FocusTracking(Enum):
    """How closely flat review follows the focus/caret location."""

    OFF = 0
    AUTO = 1
    ON = 2


@gsettings_registry.get_registry().gsettings_schema("org.gnome.Orca.FlatReview", name="flat-review")
class FlatReviewPresenter(Extension):
    """Provides access to on-screen objects via flat-review."""

    _SCHEMA = "flat-review"
    KEY_DISPLAY_UPDATES = "display-updates"
    KEY_FOCUS_TRACKING = "focus-tracking"
    KEY_RESTRICTED = "restricted"
    KEY_SPEAK_UPDATES = "speak-updates"

    def _get_setting(self, key: str, default: bool) -> bool:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            "b",
            default=default,
        )

    GROUP_LABEL = guilabels.KB_GROUP_FLAT_REVIEW

    _EVENT_TYPES: tuple[str, ...] = ("object:text-changed:insert", "object:text-caret-moved")

    def __init__(self) -> None:
        self._context: flat_review.Context | None = None
        self._current_contents: str = ""
        self._last_presented_unit: int = flat_review.Context.LINE
        self._last_input_event: input_event.InputEvent | None = None
        self._context_input_event: input_event.InputEvent | None = None
        self._restrict: bool = self.get_is_restricted()
        self._gui: FlatReviewContextGUI | None = None
        self._event_listener: Atspi.EventListener = Atspi.EventListener.new(self._listener)
        self._registered_app: Atspi.Accessible | None = None
        self._context_invalidated: bool = False
        self._location_invalidated: bool = False
        self._pending_caret_moved: Atspi.Event | None = None
        self._pending_text_inserted: Atspi.Event | None = None
        self._idle_id: int = 0
        self._region_changed_listener_registered: bool = False
        super().__init__()

    def _listener(self, event: Atspi.Event) -> None:
        """Generic event listener."""

        # TODO - JD: Implement support to invalidate individual objects.
        if self._context is None:
            return

        if event.source != self._context.get_current_object():
            return

        if event.type == "object:text-changed:insert":
            self._pending_text_inserted = event
        elif event.type == "object:text-caret-moved":
            self._pending_caret_moved = event

        if self._idle_id == 0:
            self._idle_id = GLib.idle_add(self._process_pending_events)

    def _change_invalidates_location(
        self, text_inserted: Atspi.Event | None, caret_moved: Atspi.Event | None
    ) -> bool:
        """Returns True if a change to the reviewed object should move the review location."""

        tracking = self.get_focus_tracking()
        if tracking == FocusTracking.OFF.value:
            return False
        if tracking == FocusTracking.ON.value:
            if (
                caret_moved is not None
                and AXUtilities.get_text_event_reason(caret_moved)
                != TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            ):
                return True
            return (
                text_inserted is not None
                and AXUtilities.get_text_event_reason(text_inserted)
                != TextEventReason.AUTO_INSERTION_UNPRESENTABLE
            )

        if text_inserted is not None and AXUtilities.get_text_event_reason(text_inserted) in (
            TextEventReason.TYPING,
            TextEventReason.AUTO_INSERTION_PRESENTABLE,
        ):
            return True
        return caret_moved is not None and AXUtilities.get_text_event_reason(caret_moved) in (
            TextEventReason.NAVIGATION_BY_LINE,
            TextEventReason.NAVIGATION_TO_FILE_BOUNDARY,
        )

    def _process_pending_events(self) -> bool:
        """Processes pending events. Returns False so GLib doesn't reschedule."""

        self._idle_id = 0
        caret_moved = self._pending_caret_moved
        text_inserted = self._pending_text_inserted
        self._pending_caret_moved = None
        self._pending_text_inserted = None

        if self._context is None:
            return False

        current_obj = self._context.get_current_object()
        if caret_moved is not None and caret_moved.source != current_obj:
            caret_moved = None
        if text_inserted is not None and text_inserted.source != current_obj:
            text_inserted = None

        if caret_moved is None and text_inserted is None:
            return False

        self._context_invalidated = True
        if self._change_invalidates_location(text_inserted, caret_moved):
            self._location_invalidated = True

        if not self.last_input_event_was_review_command():
            msg = (
                "FLAT REVIEW PRESENTER: Not refreshing braille; "
                "last input event was not a review command."
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        script = script_manager.get_manager().get_active_script()
        if script is not None:
            self._present_review_update(script)

        return False

    def _register_event_listeners(self, app: Atspi.Accessible) -> None:
        """Registers the event listeners for app, replacing any previous registration."""

        if app == self._registered_app:
            return
        if self._registered_app is not None:
            self._deregister_event_listeners()
        for event_type in self._EVENT_TYPES:
            self._event_listener.register_with_app(event_type, app=app)
        self._registered_app = app

    def _deregister_event_listeners(self) -> None:
        """Deregisters the event listeners if currently registered."""

        if self._registered_app is None:
            return

        try:
            for event_type in self._EVENT_TYPES:
                self._event_listener.deregister(event_type)
        except GLib.GError as error:
            msg = f"FLAT REVIEW PRESENTER: Exception deregistering {event_type}: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)

        self._registered_app = None

    def _get_commands(self) -> list[Command]:
        return flat_review_presenter_command_definitions.get_commands(self)

    def is_active(self) -> bool:
        """Returns True if the flat review presenter is active."""

        mode, _obj = focus_manager.get_manager().get_active_mode_and_object_of_interest()
        return mode == focus_manager.FLAT_REVIEW

    def last_input_event_was_review_command(self) -> bool:
        """Returns True if the last input event was a flat review command."""

        if self._last_input_event is None:
            return False

        manager = input_event_manager.get_manager()
        result = manager.previous_event_equals_or_is_release_for_event(self._last_input_event)
        string = self._last_input_event.as_single_line_string()

        msg = f"FLAT REVIEW PRESENTER: Last review event ({string}) is last input event: {result}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    def _can_use_existing_context(  # pylint: disable=too-many-return-statements
        self, is_panning: bool = False
    ) -> bool:
        """Returns True if the existing context can be used."""

        if not self._context:
            return False

        if self._context_invalidated:
            msg = "FLAT REVIEW PRESENTER: Context invalidated by an accessible event."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        last_input_event = input_event_manager.get_manager().get_last_input_event()
        if last_input_event is not None and last_input_event is self._context_input_event:
            msg = "FLAT REVIEW PRESENTER: No new input event since context was last validated."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self._context.is_stale():
            msg = "FLAT REVIEW PRESENTER: Context is stale."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if self._last_input_event is None:
            return True

        if is_panning:
            return True

        # TODO - JD: Add some additional logic for cases where we know that the previous input
        # event should not invalidate the context. For instance, the system information presenter
        # and where am I presenter will never trigger scrolling, text changes, etc.

        return self.last_input_event_was_review_command()

    # pylint: disable-next=too-many-locals
    def get_or_create_context(
        self, script: default.Script | None = None, is_panning: bool = False
    ) -> flat_review.Context:
        """Returns the flat review context, creating one if necessary."""

        if not self._can_use_existing_context(is_panning):
            msg = f"FLAT REVIEW PRESENTER: Creating new context. Restrict: {self._restrict}"
            debug.print_message(debug.LEVEL_INFO, msg, True)

            previous_obj, previous_location = None, None
            if self._context is not None and not self._location_invalidated:
                previous_obj = self._context.get_current_object()
                previous_location = self._context.get_current_location()

            manager = focus_manager.get_manager()
            if self._restrict:
                mode, obj = manager.get_active_mode_and_object_of_interest()
                self._context = flat_review.Context(script, root=obj)
            else:
                self._context = flat_review.Context(script)

            # The new context is built at the focus, so move it back onto the reviewed object.
            if (
                not self._restrict
                and previous_obj is not None
                and self.get_focus_tracking() == FocusTracking.OFF.value
            ):
                self._context.set_current_to_zone_with_object(previous_obj)

            current_obj = self._context.get_current_object()
            if previous_obj == current_obj and previous_location is not None:
                line, zone, word, char = previous_location
                # Try the exact location first; if the new context's structure has
                # shifted (e.g. a word got shorter), fall back progressively so we
                # still land on the right line.
                for fallback in (
                    (line, zone, word, char),
                    (line, zone, word, 0),
                    (line, zone, 0, 0),
                    (line, 0, 0, 0),
                ):
                    if self._context.can_set_location(fallback):
                        tokens = [
                            "FLAT REVIEW PRESENTER: Restoring previous location in new context:",
                            current_obj,
                            fallback,
                        ]
                        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                        self._context.set_current_location(fallback)
                        break

            manager.emit_region_changed(current_obj, mode=focus_manager.FLAT_REVIEW)
            self._context_input_event = input_event_manager.get_manager().get_last_input_event()
            app = AXUtilities.get_application(current_obj)
            if app is not None:
                self._register_event_listeners(app)
            self._context_invalidated = False
            self._location_invalidated = False
            self._sync_region_changed_listener()
            return self._context

        assert self._context is not None
        msg = f"FLAT REVIEW PRESENTER: Using existing context. Restrict: {self._restrict}"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        # If we are in unrestricted mode, update the context as below.
        # If the context already exists, but the active mode is not flat review, update
        # the flat review location to that of the object of interest -- if the object of
        # interest is in the flat review context (which means it's on screen). In some
        # cases the object of interest will not be in the flat review context because it
        # is represented by descendant text objects. set_current_to_zone_with_object checks
        # for this condition and if it can find a zone whose ancestor is the object of
        # interest, it will set the current zone to the descendant, causing Orca to
        # present the text at the location of the object of interest.
        mode, obj = focus_manager.get_manager().get_active_mode_and_object_of_interest()
        obj = obj or focus_manager.get_manager().get_locus_of_focus()
        # On invalidates the context elsewhere; off stays put; only auto moves the location here.
        if (
            self.get_focus_tracking() == FocusTracking.AUTO.value
            and mode != focus_manager.FLAT_REVIEW
            and obj != self._context.get_current_object()
            and not self._restrict
        ):
            tokens = [
                "FLAT REVIEW PRESENTER: Attempting to update location from",
                self._context.get_current_object(),
                "to",
                obj,
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            self._context.set_current_to_zone_with_object(obj)

        # If we are restricting, and the current mode is not flat review, calculate a new context
        if self._restrict and mode != focus_manager.FLAT_REVIEW:
            msg = "FLAT REVIEW PRESENTER: Creating new restricted context."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._context = flat_review.Context(script, obj)

        self._context_input_event = input_event_manager.get_manager().get_last_input_event()
        return self._context

    def start(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> None:
        """Starts flat review."""

        if self.is_active():
            msg = "FLAT REVIEW PRESENTER: Already in flat review"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        msg = "FLAT REVIEW PRESENTER: Starting flat review"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        if script is None:
            script = script_manager.get_manager().get_active_script()
            assert script is not None

        context = self.get_or_create_context(script)
        manager = focus_manager.get_manager()
        if manager.get_active_mode_and_object_of_interest()[0] != focus_manager.FLAT_REVIEW:
            manager.emit_region_changed(
                context.get_current_object(), mode=focus_manager.FLAT_REVIEW
            )
        if event is None:
            return

        if speech_presenter.get_presenter().use_verbose_speech():
            presentation_manager.get_manager().present_message(messages.FLAT_REVIEW_START)
        self._item_presentation(script, event)

    def quit(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> None:
        """Quits flat review."""

        if self._context is None:
            msg = "FLAT REVIEW PRESENTER: Not in flat review"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        msg = "FLAT REVIEW PRESENTER: Quitting flat review"
        debug.print_message(debug.LEVEL_INFO, msg, True)

        self._last_input_event = None
        self._context = None
        self._context_input_event = None
        self._deregister_event_listeners()
        self._sync_region_changed_listener()
        self._context_invalidated = False
        self._location_invalidated = False
        if self._idle_id:
            GLib.source_remove(self._idle_id)
            self._idle_id = 0
        self._pending_caret_moved = None
        self._pending_text_inserted = None
        focus = focus_manager.get_manager().get_locus_of_focus()
        focus_manager.get_manager().emit_region_changed(focus, mode=focus_manager.FOCUS_TRACKING)
        if event is None or script is None:
            return

        if speech_presenter.get_presenter().use_verbose_speech():
            presentation_manager.get_manager().present_message(messages.FLAT_REVIEW_STOP)
        script.update_braille(focus)

    def _sync_region_changed_listener(self) -> None:
        """Listens for focus changes only while a review context exists."""

        needed = self._context is not None
        manager = focus_manager.get_manager()
        if needed and not self._region_changed_listener_registered:
            manager.add_region_changed_listener(self._on_region_changed)
            self._region_changed_listener_registered = True
        elif not needed and self._region_changed_listener_registered:
            manager.remove_region_changed_listener(self._on_region_changed)
            self._region_changed_listener_registered = False

    def _on_region_changed(self, _obj: Atspi.Accessible | None, mode: str) -> None:
        """Resets the review state when focus leaves flat review and tracking is on."""

        if mode == focus_manager.FLAT_REVIEW:
            return
        if self.get_focus_tracking() != FocusTracking.ON.value:
            return

        msg = "FLAT REVIEW PRESENTER: Focus left flat review; resetting context."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._context_invalidated = True
        self._location_invalidated = True
        self._last_input_event = None

    @dbus_service.command
    def toggle_flat_review_mode(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = False,
    ) -> bool:
        """Toggles between flat review mode and focus tracking mode."""

        tokens = [
            "FLAT REVIEW PRESENTER: toggle_flat_review_mode. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if self.is_active():
            self.quit(script, event)
            return True

        self.start(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_home(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,  # pylint: disable=unused-argument
    ) -> bool:
        """Moves to the top left of the current window."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_home. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_start_of(flat_review.Context.WINDOW)
        self.present_line(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_end(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the bottom right of the current window."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_end. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_end_of(flat_review.Context.WINDOW)
        self.present_line(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_bottom_left(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the bottom left of the current window."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_bottom_left. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_end_of(flat_review.Context.WINDOW)
        self._context.go_to_start_of(flat_review.Context.LINE)
        self.present_line(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_previous_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the previous line."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_previous_line. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_previous_line():
            self.present_line(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def present_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current line."""

        tokens = [
            "FLAT REVIEW PRESENTER: present_line. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._line_presentation(script, event, 1)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_next_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the next line."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_next_line. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_next_line():
            self.present_line(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def spell_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current line letter by letter."""

        tokens = [
            "FLAT REVIEW PRESENTER: spell_line. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._line_presentation(script, event, 2)
        self._last_input_event = event
        return True

    @dbus_service.command
    def phonetic_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current line letter by letter phonetically."""

        tokens = [
            "FLAT REVIEW PRESENTER: phonetic_line. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._line_presentation(script, event, 3)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_start_of_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the beginning of the current line."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_start_of_line. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_start_of(flat_review.Context.LINE)
        self.present_character(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_end_of_line(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the end of the line."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_end_of_line. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._context.go_to_end_of(flat_review.Context.LINE)
        self.present_character(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_previous_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the previous item or word."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_previous_item. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_previous_word():
            self.present_item(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def present_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current item/word."""

        tokens = [
            "FLAT REVIEW PRESENTER: present_item. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._item_presentation(script, event, 1)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_next_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the next item or word."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_next_item. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_next_word():
            self.present_item(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def spell_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current item/word letter by letter."""

        tokens = [
            "FLAT REVIEW PRESENTER: spell_item. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._item_presentation(script, event, 2)
        self._last_input_event = event
        return True

    @dbus_service.command
    def phonetic_item(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current word letter by letter phonetically."""

        tokens = [
            "FLAT REVIEW PRESENTER: phonetic_item. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._item_presentation(script, event, 3)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_previous_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the previous character."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_previous_character. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_previous_character():
            self.present_character(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def present_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current character."""

        tokens = [
            "FLAT REVIEW PRESENTER: present_character. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._character_presentation(script, event, 1)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_next_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the next character."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_next_character. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_next_character():
            self.present_character(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def spell_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current character phonetically."""

        tokens = [
            "FLAT REVIEW PRESENTER: spell_character. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._character_presentation(script, event, 2)
        self._last_input_event = event
        return True

    @dbus_service.command
    def unicode_current_character(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current character's unicode value."""

        tokens = [
            "FLAT REVIEW PRESENTER: unicode_current_character. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._character_presentation(script, event, 3)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_above(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the character above."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_above. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_up():
            self.present_item(script, event)
        self._last_input_event = event
        return True

    @dbus_service.command
    def go_below(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves to the character below."""

        tokens = [
            "FLAT REVIEW PRESENTER: go_below. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if self._context.go_down():
            self.present_item(script, event)
        self._last_input_event = event
        return True

    def get_current_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> Atspi.Accessible:
        """Returns the current accessible object."""

        tokens = [
            "FLAT REVIEW PRESENTER: get_current_object. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._last_input_event = event
        return self._context.get_current_object()

    @dbus_service.command
    def move_review_to_focus(  # pylint: disable=unused-argument
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Moves the flat review location to the object with focus."""

        tokens = ["FLAT REVIEW PRESENTER: move_review_to_focus. Script:", script, "Event:", event]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._last_input_event = event
        context = self.get_or_create_context(script)
        obj = focus_manager.get_manager().get_locus_of_focus()
        if not context.set_current_to_zone_with_object(obj):
            self._context = None
            self.get_or_create_context(script)
        return self._item_presentation(script, event)

    @dbus_service.command
    def move_focus_to_review(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Attempts to move focus to the object at the flat review location."""

        tokens = ["FLAT REVIEW PRESENTER: move_focus_to_review. Script:", script, "Event:", event]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._last_input_event = event
        obj = self._context.get_current_object()

        # If focus is already on the reviewed object, no focus event will present the change,
        # so capture how far the review location is from the caret to announce the move.
        focus_stays = obj == focus_manager.get_manager().get_locus_of_focus()
        old_location = target_location = self._context.get_current_location()
        if focus_stays:
            self._context.set_current_to_zone_with_object(obj)
            old_location = self._context.get_current_location()
            self._context.set_current_location(target_location)

        AXObject.grab_focus(obj)
        offset = self._context.get_current_text_offset()
        caret_moved = False
        if offset >= 0:
            AXUtilities.set_caret_offset_with_reason(obj, offset, CaretSetReason.FLAT_REVIEW)
            caret_moved = AXText.get_caret_offset(obj) == offset

        if AXUtilities.is_focusable(obj):
            if focus_stays and not caret_moved:
                if notify_user:
                    presentation_manager.get_manager().present_message(
                        messages.FLAT_REVIEW_LOCATION_UNCHANGED
                    )
            elif focus_stays:
                if old_location[0] != target_location[0]:
                    self._line_presentation(script, event)
                elif old_location[1:3] != target_location[1:3]:
                    self._item_presentation(script, event)
                else:
                    self._character_presentation(script, event)
            return True
        if script.utilities.in_document_content(obj):
            focus_manager.get_manager().set_locus_of_focus(None, obj)
        elif notify_user:
            presentation_manager.get_manager().present_message(
                messages.FLAT_REVIEW_LOCATION_UNCHANGED
            )
        return True

    @dbus_service.command
    def present_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the current accessible object."""

        tokens = [
            "FLAT REVIEW PRESENTER: present_object. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        if not isinstance(event, input_event.BrailleEvent):
            script.present_object(self._context.get_current_object(), generate_braille=False)

        focus_manager.get_manager().emit_region_changed(
            self._context.get_current_object(),
            mode=focus_manager.FLAT_REVIEW,
        )
        self._last_input_event = event
        return True

    @dbus_service.command
    def left_click_on_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Attempts to synthesize a left click on the current accessible."""

        tokens = [
            "FLAT REVIEW PRESENTER: left_click_on_object. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._last_input_event = event
        obj = self._context.get_current_object()
        offset = self._context.get_current_text_offset()
        if offset >= 0 and AXEventSynthesizer.click_character(obj, offset, 1):
            return True
        return AXEventSynthesizer.click_object(obj, 1)

    @dbus_service.command
    def right_click_on_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Attempts to synthesize a right click on the current accessible."""

        tokens = [
            "FLAT REVIEW PRESENTER: right_click_on_object. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._last_input_event = event
        obj = self._context.get_current_object()
        offset = self._context.get_current_text_offset()
        if offset >= 0 and AXEventSynthesizer.click_character(obj, offset, 3):
            return True
        return AXEventSynthesizer.click_object(obj, 3)

    @dbus_service.command
    def route_pointer_to_object(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Routes the mouse pointer to the current accessible."""

        tokens = [
            "FLAT REVIEW PRESENTER: route_pointer_to_object. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._context = self.get_or_create_context(script)
        self._last_input_event = event
        obj = self._context.get_current_object()
        offset = self._context.get_current_text_offset()
        if offset >= 0 and AXEventSynthesizer.route_to_character(obj, offset):
            return True
        return AXEventSynthesizer.route_to_object(obj)

    def _update_braille(self, script: default.Script) -> None:
        """Obtains the braille regions for the current flat review line and displays them."""

        if not braille_presenter.get_presenter().use_braille():
            return

        self._context = self.get_or_create_context(script)
        regions, focused_region = self._context.get_current_braille_regions()
        if not regions:
            braille_presenter.get_presenter().refresh_braille()
            return

        braille_presenter.get_presenter().present_regions(
            list(regions),
            focused_region,
        )

    def _speak_review_unit(self, string: str) -> None:
        """Speaks string as the current review unit (used for live updates)."""

        presenter = presentation_manager.get_manager()
        if not string or string == "\n":
            presenter.speak_message(messages.BLANK)
        elif string.isspace():
            presenter.speak_message(messages.WHITE_SPACE)
        else:
            obj = self._context.get_current_object() if self._context else None
            presenter.speak_accessible_text(obj, string)

    def _present_review_update(self, script: default.Script) -> None:
        """Presents the current review unit when it changes via a live update."""

        self._context = self.get_or_create_context(script)
        if self.get_displays_updates():
            self._update_braille(script)

        if not self.get_speaks_updates():
            return

        if self._last_presented_unit == flat_review.Context.WORD:
            current = self._context.get_current_word_string()
        elif self._last_presented_unit == flat_review.Context.CHAR:
            current = self._context.get_current_character_string()
        else:
            current = self._context.get_current_line_string()

        if current == self._current_contents:
            return

        self._speak_review_unit(current)
        self._current_contents = current

    def pan_braille_left(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Pans the braille display left."""

        self._context = self.get_or_create_context(script, is_panning=True)
        self._last_input_event = event

        # Try to pan left. If we couldn't (at edge), move to previous line.
        if not braille_presenter.get_presenter().pan_left():
            self._context.go_to_start_of(flat_review.Context.LINE)
            self._context.go_previous_character()
            self._update_braille(script)
            return True

        # After panning, find a region that has a zone attribute to sync flat review context.
        # We do cell-by-cell panning here because not all braille regions have zones (e.g.,
        # buttons don't have zones, only text regions do). This finds the first region with
        # a zone so we can update the flat review position accordingly.
        region_info = braille.get_region_at_cell(1)
        braille_region = region_info.region
        offset_in_zone = region_info.offset_in_region
        while braille_region and not hasattr(braille_region, "zone"):
            # Cell-by-cell panning (amount=1) to find the first region with a zone attribute.
            if not braille.pan_left(1):
                break
            region_info = braille.get_region_at_cell(1)
            braille_region = region_info.region
            offset_in_zone = region_info.offset_in_region

        if braille_region and not hasattr(braille_region, "zone"):
            self._context.go_to_start_of(flat_review.Context.LINE)
            self._context.go_previous_character()
            self._update_braille(script)
        elif braille_region and hasattr(braille_region, "zone"):
            self._context.set_current_zone(braille_region.zone, offset_in_zone)
            braille_presenter.get_presenter().refresh_braille(pan_to_cursor=False)

        return True

    def pan_braille_right(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Pans the braille display right."""

        self._context = self.get_or_create_context(script, is_panning=True)
        self._last_input_event = event

        # Try to pan right. If we couldn't (at edge), move to next line.
        if not braille_presenter.get_presenter().pan_right():
            self._context.go_next_line()
            self._update_braille(script)
            return True

        # After panning, find a region that has a zone attribute to sync flat review context.
        # We do cell-by-cell panning here because not all braille regions have zones (e.g.,
        # buttons don't have zones, only text regions do). This finds the first region with
        # a zone so we can update the flat review position accordingly.
        region_info = braille.get_region_at_cell(1)
        braille_region = region_info.region
        offset_in_zone = region_info.offset_in_region
        while braille_region and not hasattr(braille_region, "zone"):
            # Cell-by-cell panning (amount=1) to find the first region with a zone attribute.
            if not braille.pan_right(1):
                break
            region_info = braille.get_region_at_cell(1)
            braille_region = region_info.region
            offset_in_zone = region_info.offset_in_region

        if braille_region and not hasattr(braille_region, "zone"):
            self._context.go_next_line()
            self._update_braille(script)
        elif braille_region and hasattr(braille_region, "zone"):
            self._context.set_current_zone(braille_region.zone, offset_in_zone)
            braille_presenter.get_presenter().refresh_braille(pan_to_cursor=False)

        return True

    def _get_all_lines(
        self,
        script: default.Script,
        _event: input_event.InputEvent | None = None,
    ) -> tuple[list[tuple[str, Atspi.Accessible | None]], tuple[int, int, int, int]]:
        """Returns a (list of (line_string, obj) tuples, current location) tuple."""

        lines: list[tuple[str, Atspi.Accessible | None]] = []
        self._context = self.get_or_create_context(script)
        location = self._context.get_current_location()
        self._context.go_to_start_of(flat_review.Context.WINDOW)
        string = self._context.get_current_line_string()
        while string is not None:
            obj = self._context.get_current_object()
            lines.append((string.rstrip("\n"), obj))
            if not self._context.go_next_line():
                break
            string = self._context.get_current_line_string()
        return lines, location

    @dbus_service.command
    def say_all(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Speaks the contents of the entire window."""

        tokens = [
            "FLAT REVIEW PRESENTER: say_all. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        for string, obj in self._get_all_lines(script, event)[0]:
            if not string.isspace():
                presentation_manager.get_manager().speak_accessible_text(obj, string)

        self._last_input_event = event
        return True

    @dbus_service.command
    def show_contents(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Displays the entire flat review contents in a text view."""

        tokens = [
            "FLAT REVIEW PRESENTER: show_contents. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        line_tuples, location = self._get_all_lines(script, event)
        text = "\n".join(string for string, _obj in line_tuples)
        title = guilabels.FLAT_REVIEW_CONTENTS
        self._gui = FlatReviewContextGUI(script, title, text, location)
        self._gui.show_gui()
        self._last_input_event = event
        return True

    @dbus_service.command
    def copy_to_clipboard(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Copies the string just presented to the clipboard."""

        tokens = [
            "FLAT REVIEW PRESENTER: copy_to_clipboard. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not self.is_active():
            if notify_user:
                presentation_manager.get_manager().present_message(messages.FLAT_REVIEW_NOT_IN)
            return True

        clipboard.get_presenter().set_text(self._current_contents.rstrip("\n"))
        if notify_user:
            presentation_manager.get_manager().present_message(messages.FLAT_REVIEW_COPIED)
        self._last_input_event = event
        return True

    @dbus_service.command
    def append_to_clipboard(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Appends the string just presented to the clipboard."""

        tokens = [
            "FLAT REVIEW PRESENTER: append_to_clipboard. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not self.is_active():
            if notify_user:
                presentation_manager.get_manager().present_message(messages.FLAT_REVIEW_NOT_IN)
            return True

        clipboard.get_presenter().append_text(self._current_contents.rstrip("\n"))
        if notify_user:
            presentation_manager.get_manager().present_message(messages.FLAT_REVIEW_APPENDED)
        self._last_input_event = event
        return True

    def create_preferences_grid(self) -> FlatReviewPreferencesGrid:
        """Returns the GtkGrid containing the Flat Review preferences UI."""

        # pylint: disable-next=import-outside-toplevel
        from .flat_review_presenter_preferences_grid import FlatReviewPreferencesGrid

        return FlatReviewPreferencesGrid(self)

    @gsettings_registry.get_registry().gsetting(
        key=KEY_RESTRICTED,
        schema="flat-review",
        gtype="b",
        default=False,
        summary="Restrict flat review to current object",
        migration_key="flatReviewIsRestricted",
    )
    @dbus_service.getter
    def get_is_restricted(self) -> bool:
        """Returns whether flat review is restricted to the current object."""

        return self._get_setting(self.KEY_RESTRICTED, False)

    @dbus_service.setter
    def set_is_restricted(self, value: bool) -> bool:
        """Sets whether flat review is restricted to the current object."""

        msg = f"FLAT REVIEW PRESENTER: Setting is-restricted to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_RESTRICTED, value
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SPEAK_UPDATES,
        schema="flat-review",
        gtype="b",
        default=False,
        summary="Speak content changes at the flat review location",
    )
    @dbus_service.getter
    def get_speaks_updates(self) -> bool:
        """Returns whether changes at the review location are spoken automatically."""

        return self._get_setting(self.KEY_SPEAK_UPDATES, False)

    @dbus_service.setter
    def set_speaks_updates(self, value: bool) -> bool:
        """Sets whether changes at the review location are spoken automatically."""

        msg = f"FLAT REVIEW PRESENTER: Setting speak-updates to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_SPEAK_UPDATES, value
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_DISPLAY_UPDATES,
        schema="flat-review",
        gtype="b",
        default=True,
        summary="Refresh braille for content changes at the flat review location",
    )
    @dbus_service.getter
    def get_displays_updates(self) -> bool:
        """Returns whether changes at the review location refresh braille automatically."""

        return self._get_setting(self.KEY_DISPLAY_UPDATES, True)

    @dbus_service.setter
    def set_displays_updates(self, value: bool) -> bool:
        """Sets whether changes at the review location refresh braille automatically."""

        msg = f"FLAT REVIEW PRESENTER: Setting display-updates to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_DISPLAY_UPDATES, value
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FOCUS_TRACKING,
        schema="flat-review",
        genum="org.gnome.Orca.FocusTracking",
        default="auto",
        summary="When flat review should follow the focus/caret location",
    )
    @dbus_service.getter
    def get_focus_tracking(self) -> UInt32:
        """Returns the flat review focus-tracking mode as a FocusTracking value."""

        nick = gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_FOCUS_TRACKING,
            "",
            genum="org.gnome.Orca.FocusTracking",
            default="auto",
        )
        values = gsettings_registry.get_registry().get_enum_values("org.gnome.Orca.FocusTracking")
        if values and nick in values:
            return values[nick]
        return FocusTracking.AUTO.value

    @dbus_service.setter
    def set_focus_tracking(self, value: UInt32) -> bool:
        """Sets the flat review focus-tracking mode."""

        msg = f"FLAT REVIEW PRESENTER: Setting focus-tracking to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        mode = FocusTracking(value)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_FOCUS_TRACKING, mode.name.lower()
        )
        return True

    @dbus_service.command
    def toggle_restrict(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Toggles the restricting of flat review to the current object."""

        tokens = [
            "FLAT REVIEW PRESENTER: toggle_restrict. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        self._restrict = not self._restrict
        self.set_is_restricted(self._restrict)

        if self._restrict:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.FLAT_REVIEW_RESTRICTED)
        elif notify_user:
            presentation_manager.get_manager().present_message(messages.FLAT_REVIEW_UNRESTRICTED)
        if self.is_active():
            # Reset the context
            self._context = None
            self.start()

        self._last_input_event = event
        return True

    def _line_presentation(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        speech_type: int = 1,
    ) -> bool:
        """Presents the current line."""

        self._context = self.get_or_create_context(script)
        line_string = self._context.get_current_line_string()

        if not isinstance(event, input_event.BrailleEvent):
            presenter = presentation_manager.get_manager()
            if not line_string or line_string == "\n":
                presenter.speak_message(messages.BLANK)
            elif line_string.isspace():
                presenter.speak_message(messages.WHITE_SPACE)
            else:
                indentation = speech_presenter.get_presenter().get_indentation_description(
                    line_string
                )
                if indentation:
                    presenter.speak_message(indentation)
                if line_string.isupper() and (speech_type < 2 or speech_type > 3):
                    presenter.speak_accessible_text(
                        self._context.get_current_object(),
                        line_string,
                        self._context.get_current_line_start_offset(),
                    )
                elif speech_type == 2:
                    presenter.spell_item(
                        line_string,
                        self._context.get_current_object(),
                        self._context.get_current_line_start_offset(),
                    )
                elif speech_type == 3:
                    presenter.spell_phonetically(
                        line_string,
                        self._context.get_current_object(),
                        self._context.get_current_line_start_offset(),
                    )
                else:
                    presenter.speak_accessible_text(
                        self._context.get_current_object(),
                        line_string,
                        self._context.get_current_line_start_offset(),
                    )

        focus_manager.get_manager().emit_region_changed(
            self._context.get_current_object(),
            mode=focus_manager.FLAT_REVIEW,
        )
        self._update_braille(script)
        self._last_presented_unit = flat_review.Context.LINE
        self._current_contents = line_string
        return True

    def _item_presentation(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        speech_type: int = 1,
    ) -> bool:
        """Presents the current item/word."""

        self._context = self.get_or_create_context(script)
        word_string = self._context.get_current_word_string()
        if not isinstance(event, input_event.BrailleEvent):
            presenter = presentation_manager.get_manager()
            if not word_string or word_string == "\n":
                presenter.speak_message(messages.BLANK)
            else:
                line_string = self._context.get_current_line_string()
                if line_string == "\n":
                    presenter.speak_message(messages.BLANK)
                elif word_string.isspace():
                    presenter.speak_message(messages.WHITE_SPACE)
                elif speech_type == 2:
                    presenter.spell_item(
                        word_string,
                        self._context.get_current_object(),
                        self._context.get_current_word_start_offset(),
                    )
                elif speech_type == 3:
                    presenter.spell_phonetically(
                        word_string,
                        self._context.get_current_object(),
                        self._context.get_current_word_start_offset(),
                    )
                elif (
                    speech_type == 1
                    and (offset := self._context.get_current_word_start_offset()) is not None
                ):
                    speech_presenter.get_presenter().speak_phrase(
                        script,
                        self._context.get_current_object(),
                        offset,
                        offset + len(word_string),
                        word_string,
                    )

        focus_manager.get_manager().emit_region_changed(
            self._context.get_current_object(),
            mode=focus_manager.FLAT_REVIEW,
        )
        self._update_braille(script)
        self._last_presented_unit = flat_review.Context.WORD
        self._current_contents = word_string
        return True

    def _character_presentation(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        speech_type: int = 1,
    ) -> bool:
        """Presents the current character."""

        focus = focus_manager.get_manager().get_locus_of_focus()
        if not self._context and AXObject.supports_text(focus):
            char_string, offset, _end_offset = AXText.get_character_at_offset(focus)
            obj = focus
        else:
            self._context = self.get_or_create_context(script)
            char_string = self._context.get_current_character_string()
            obj = self._context.get_current_object()
            offset = self._context.get_current_text_offset()
        if not isinstance(event, input_event.BrailleEvent):
            presenter = presentation_manager.get_manager()
            if not char_string or (char_string == "\n" and speech_type != 3):
                presenter.speak_message(messages.BLANK)
            elif speech_type == 3:
                presenter.speak_message(messages.UNICODE % f"{ord(char_string):04x}")
            elif speech_type == 2:
                presenter.spell_phonetically(char_string, obj, offset if offset >= 0 else None)
            elif obj is not None and offset >= 0:
                presenter.speak_character_at_offset(obj, offset, char_string)
            else:
                presenter.speak_character(char_string)

        if not self._context:
            return True

        focus_manager.get_manager().emit_region_changed(
            self._context.get_current_object(),
            mode=focus_manager.FLAT_REVIEW,
        )
        self._update_braille(script)
        self._last_presented_unit = flat_review.Context.CHAR
        self._current_contents = char_string
        return True


class FlatReviewContextGUI:
    """Presents the entire flat review context in a text view"""

    def __init__(
        self,
        script: default.Script,
        title: str,
        text: str,
        location: tuple[int, int, int, int],
    ) -> None:
        self._script: default.Script = script
        self._gui: Gtk.Dialog = self._create_dialog(title, text, location)

    def _create_dialog(
        self,
        title: str,
        text: str,
        location: tuple[int, int, int, int],
    ) -> Gtk.Dialog:
        """Creates the dialog."""

        dialog = Gtk.Dialog(
            title,
            None,
            Gtk.DialogFlags.MODAL,
            (Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE),
        )
        dialog.set_default_size(800, 600)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_hexpand(True)
        scrolled_window.set_vexpand(True)

        textbuffer = Gtk.TextBuffer()
        textbuffer.set_text(text)

        line_index, _zone_index, word_index, char_index = location
        start_iter = textbuffer.get_start_iter()
        start_iter.forward_lines(line_index)
        for _ in range(word_index):
            start_iter.forward_word_end()
            # This is needed to get past any punctuation. Note that we cannot check starts_word
            # here. Example: The letter after an apostrophe is reported as a word start.
            while not start_iter.get_char().isspace():
                if not start_iter.forward_char():
                    break

        # The text iter word can start with one or more spaces. Move to the beginning of the flat
        # review word before advancing by character.
        while not start_iter.starts_word():
            if not start_iter.forward_char():
                break

        start_iter.forward_chars(char_index)
        textbuffer.place_cursor(start_iter)

        textview = Gtk.TextView(buffer=textbuffer)
        textview.set_wrap_mode(Gtk.WrapMode.WORD)
        scrolled_window.add(textview)  # pylint: disable=no-member
        dialog.get_content_area().pack_start(scrolled_window, True, True, 0)
        dialog.connect("response", self.on_response)
        return dialog

    def on_response(self, _dialog: Gtk.Dialog, response: Gtk.ResponseType) -> None:
        """Handler for the 'response' signal of the dialog."""

        if response == Gtk.ResponseType.CLOSE:
            self._gui.destroy()

    def show_gui(self) -> None:
        """Shows the dialog."""

        self._gui.show_all()  # pylint: disable=no-member
        self._gui.present_with_time(time.time())


_presenter: FlatReviewPresenter = FlatReviewPresenter()


def get_presenter() -> FlatReviewPresenter:
    """Returns the Flat Review Presenter"""

    return _presenter



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2010-2011 Orca Team
# Copyright 2011-2015 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-locals

"""Produces braille presentation for accessible objects."""

from __future__ import annotations

from dataclasses import replace
from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from orca import braille, braille_generator, debug, messages, object_properties
from orca.ax_object import AXObject
from orca.ax_utilities import AXUtilities
from orca.generator import ContentItem, ContentPosition

if TYPE_CHECKING:
    from orca.braille_generator import BrailleGeneratorContext

    from . import script


class BrailleGenerator(braille_generator.BrailleGenerator):
    """Produces braille presentation for accessible objects."""

    # Type annotation to override the base class script type
    _script: script.Script

    @staticmethod
    def log_generator_output(func):
        """Decorator for logging."""

        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            tokens = [f"WEB BRAILLE GENERATOR: {func.__name__}:", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return result

        return wrapper

    def get_localized_role_name(
        self, obj: Atspi.Accessible, *, role: Atspi.Role | str | None = None
    ) -> str:
        if not self._script.utilities.in_document_content(obj):
            return super().get_localized_role_name(obj, role=role)

        role_description = AXObject.get_role_description(obj, True)
        if role_description:
            return role_description

        return super().get_localized_role_name(obj, role=role)

    @log_generator_output
    def _generate_accessible_role(self, obj: Atspi.Accessible) -> list[Any]:
        """Prevents some roles from being displayed."""

        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_role(obj)

        role_description = AXObject.get_role_description(obj, True)
        if role_description:
            return [role_description]

        # TODO - JD: Can this logic be moved to the default braille generator?
        do_not_display = [
            Atspi.Role.FORM,
            Atspi.Role.PARAGRAPH,
            Atspi.Role.STATIC,
            Atspi.Role.SECTION,
            Atspi.Role.REDUNDANT_OBJECT,
            Atspi.Role.UNKNOWN,
        ]

        if not AXUtilities.is_focusable(obj):
            do_not_display.extend(
                [
                    Atspi.Role.LIST,
                    Atspi.Role.LIST_ITEM,
                    Atspi.Role.COLUMN_HEADER,
                    Atspi.Role.ROW_HEADER,
                    Atspi.Role.TABLE_CELL,
                    Atspi.Role.PANEL,
                ],
            )

        if self._get_start_offset(obj) is not None and self._get_end_offset(obj) is not None:
            do_not_display.append(Atspi.Role.ALERT)

        result = []
        role = self._get_resolved_role(obj)

        level = AXUtilities.get_heading_level(obj)
        if level:
            result.append(object_properties.ROLE_HEADING_LEVEL_BRAILLE % level)
            return result

        if self._script.utilities.is_link(obj) and obj == self._context.focus:
            if AXUtilities.is_image(AXObject.get_parent(obj)):
                result.append(messages.IMAGE_MAP_LINK)

        elif role not in do_not_display:
            label = AXUtilities.get_label_for_cell_coordinates(obj)
            if label:
                result.append(label)
            else:
                result = super()._generate_accessible_role(obj)

        if self._get_content_position(obj).index == self._get_content_position(obj).total - 1 and (
            AXUtilities.is_image(obj, self._get_resolved_role())
            or self._script.utilities.treat_as_text_object(obj)
        ):
            heading = AXUtilities.find_ancestor(obj, AXUtilities.is_heading)
            if heading is not None:
                result.extend(self._generate_accessible_role(heading))

        return result

    @log_generator_output
    def _generate_text_substring(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._script.utilities.treat_as_text_object(obj):
            return []

        return super()._generate_text_substring(obj)

    @log_generator_output
    def _generate_accessible_label(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_label(obj)

        label, _objects = self._script.utilities.infer_label_for(obj)
        if label:
            return [label]

        return super()._generate_accessible_label(obj)

    @log_generator_output
    def _generate_accessible_label_and_name(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_label_and_name(obj)

        if self._script.utilities.is_text_block_element(
            obj
        ) and self._script.utilities.treat_as_text_object(obj):
            return []

        if AXUtilities.is_editable(obj) and AXUtilities.find_ancestor(obj, AXUtilities.is_code):
            return []

        role = self._get_resolved_role(obj)
        if role == Atspi.Role.LABEL and AXObject.supports_text(obj):
            return []

        return super()._generate_accessible_label_and_name(obj)

    @log_generator_output
    def _generate_accessible_description(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_description(obj)

        # TODO - JD: Can this logic be moved into the default braille generator?
        if self._prefer_description_over_name(obj):
            return []

        return super()._generate_accessible_description(obj)

    @log_generator_output
    def _generate_accessible_name(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_name(obj)

        braille_label = AXObject.get_attributes_dict(obj).get("braillelabel")
        if braille_label:
            return [braille_label]

        # TODO - JD: Can this logic be moved into the default braille generator?
        if self._prefer_description_over_name(obj):
            return [AXObject.get_description(obj)]

        if AXObject.get_name(obj) and not self._script.utilities.has_valid_name(obj):
            return []

        result = super()._generate_accessible_name(obj)
        if result and result[0] and not AXUtilities.has_explicit_name(obj):
            result[0] = result[0].strip()
        elif not result and AXUtilities.is_check_box(obj):
            grid_cell = AXUtilities.find_ancestor(obj, AXUtilities.is_grid_cell)
            if grid_cell:
                return super()._generate_accessible_name(grid_cell)

        return result

    @log_generator_output
    def _generate_real_active_descendant_displayed_text(
        self,
        obj: Atspi.Accessible,
    ) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_real_active_descendant_displayed_text(obj)

        rad = AXUtilities.active_descendant(obj)
        return self._generate_text_content(rad)

    def generate_braille(
        self,
        obj: Atspi.Accessible,
        context: BrailleGeneratorContext,
        *,
        role: Atspi.Role | str | None = None,
        include_context: bool = True,
    ) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            tokens = ["WEB:", obj, "is not in document content. Calling default braille generator."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return super().generate_braille(
                obj, context, role=role, include_context=include_context
            )

        tokens = ["WEB: Generating braille for document object", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True, True)

        # Links in document content are presented as their own braille regions, which carry
        # their underline. Don't have the text regions scan all the document's links inline.
        context = replace(context, indicate_links=False)

        result = []

        include_context = not self._script.utilities.in_document_content(obj)
        if self._script.utilities.is_clickable_element(obj) or self._script.utilities.is_link(obj):
            role = Atspi.Role.LINK
        elif AXUtilities.is_custom_image(obj):
            role = Atspi.Role.IMAGE
        elif AXUtilities.is_anchor(obj):
            role = Atspi.Role.STATIC
        elif self._script.utilities.treat_as_div(obj, offset=self._get_start_offset(obj)):
            role = Atspi.Role.SECTION

        if AXUtilities.is_menu_item(obj):
            combo_box = AXUtilities.find_ancestor(obj, AXUtilities.is_combo_box)
            if combo_box and not AXUtilities.is_expanded(combo_box):
                obj = combo_box
        result.extend(
            super().generate_braille(obj, context, role=role, include_context=include_context),
        )
        return result

    def generate_contents(  # type: ignore[override]
        self,
        contents: list[tuple[Atspi.Accessible, int, int, str]],
        context: BrailleGeneratorContext,
    ) -> tuple[list[list[Any]], Atspi.Accessible | None]:
        self._context = context
        return self._generate_web_braille_contents(contents)

    def _generate_web_braille_contents(
        self,
        contents: list[tuple[Atspi.Accessible, int, int, str]],
    ) -> tuple[list[list[Any]], Atspi.Accessible | None]:
        if not contents:
            return [], None

        result = []
        contents = self._script.utilities.filter_contents_for_presentation(contents, True)

        obj, offset = self._script.utilities.get_caret_context()
        index = self._script.utilities.find_object_in_contents(obj, offset, contents)

        last_region = None
        focused_region = None
        original_context = self._context
        for i, content in enumerate(contents):
            acc, start, end, string = content
            item_context = replace(
                original_context,
                content_item=ContentItem(
                    start_offset=start, end_offset=end, string=string, caret_offset=offset
                ),
                content_position=ContentPosition(index=i, total=len(contents)),
                content_subject=acc,
            )
            regions, f_region = self.generate_braille(acc, item_context)
            if not regions:
                continue

            if i == index:
                focused_region = f_region

            if last_region and regions:
                last_char = next_char = ""
                if last_region.string:
                    last_char = last_region.string[-1]
                if regions[0].string:
                    next_char = regions[0].string[0]
                if self._needs_separator(last_char, next_char):
                    regions.insert(0, braille.Region(" "))

            last_region = regions[-1]
            result.append(regions)

        self._context = original_context
        return result, focused_region



# Orca
#
# Copyright 2025-2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command classes for Orca's commands."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable

    from . import input_event, keybindings
    from .scripts import default


class Command:
    """Base class for Orca commands.

    Commands have two independent activity states:

    enabled: User preference for whether this command should be active.
        - Set via user settings or toggle commands (e.g., "toggle caret navigation")
        - Persists across sessions
        - Example: User prefers caret navigation on

    suspended: Temporary system override that deactivates the command.
        - Set by Orca modes (e.g., focus mode suspends browse-mode commands)
        - Does NOT change the user's enabled preference
        - When suspension is lifted, command returns to its enabled state
        - Example: Focus mode suspends structural navigation; leaving focus
          mode automatically restores it
    """

    # pylint: disable-next=too-many-arguments, too-many-positional-arguments
    def __init__(
        self,
        name: str,
        function: Callable[..., bool],
        group_label: str,
        description: str = "",
        enabled: bool = True,
        suspended: bool = False,
        transient: bool = False,
    ) -> None:
        """Initializes a command."""

        self._name = name
        self._function = function
        self._group_label = group_label
        self._description = description
        self._enabled = enabled
        self._suspended = suspended
        self._transient = transient

    def __str__(self) -> str:
        """Returns a string representation of the command."""

        parts = [f"Command({self._name})"]
        if self._suspended:
            parts.append("SUSPENDED")
        return " ".join(parts)

    def get_name(self) -> str:
        """Returns the command name."""

        return self._name

    def get_function(self) -> Callable[..., bool]:
        """Returns the command function."""

        return self._function

    def set_function(self, function: Callable[..., bool]) -> None:
        """Sets the command function."""

        self._function = function

    def get_group_label(self) -> str:
        """Returns the group label for display grouping."""

        return self._group_label

    def get_description(self) -> str:
        """Returns the command description."""

        return self._description

    def set_group_label(self, group_label: str) -> None:
        """Sets the group label."""

        self._group_label = group_label

    def is_enabled(self) -> bool:
        """Returns True if the user has enabled this command."""

        return self._enabled

    def set_enabled(self, enabled: bool) -> None:
        """Sets whether the user has enabled this command."""

        self._enabled = enabled

    def is_suspended(self) -> bool:
        """Returns True if this command is temporarily suspended by the system."""

        return self._suspended

    def set_suspended(self, suspended: bool) -> None:
        """Sets whether this command is temporarily suspended by the system."""

        self._suspended = suspended

    def is_transient(self) -> bool:
        """Returns True if this command can come and go, so it sorts last for display."""

        return self._transient

    def set_transient(self, transient: bool) -> None:
        """Sets whether this command can come and go, so it sorts last for display."""

        self._transient = transient

    def execute(self, script: default.Script, event: input_event.InputEvent | None = None) -> bool:
        """Executes this command's function and returns True if handled."""

        return self._function(script, event)


class KeyboardCommand(Command):  # pylint: disable=too-many-instance-attributes
    """A command that can be bound to keyboard keys."""

    # pylint: disable-next=too-many-arguments, too-many-positional-arguments
    def __init__(
        self,
        name: str,
        function: Callable[..., bool],
        group_label: str,
        description: str = "",
        desktop_keybinding: keybindings.KeyBinding | None = None,
        laptop_keybinding: keybindings.KeyBinding | None = None,
        enabled: bool = True,
        suspended: bool = False,
        is_group_toggle: bool = False,
        transient: bool = False,
    ) -> None:
        """Initializes a keyboard command."""

        super().__init__(name, function, group_label, description, enabled, suspended, transient)

        # The default bindings.
        self._desktop_keybinding = desktop_keybinding
        self._laptop_keybinding = laptop_keybinding

        # The actual binding, taking into account user overrides.
        self._keybinding: keybindings.KeyBinding | None = None
        self._is_group_toggle = is_group_toggle

    def __str__(self) -> str:
        """Returns a string representation of the command."""

        parts = [f"KeyboardCommand({self._name})"]
        if self._keybinding:
            parts.append(str(self._keybinding))
        else:
            parts.append("UNBOUND")
        if self._suspended:
            parts.append("SUSPENDED")
        return " ".join(parts)

    def get_keybinding(self) -> keybindings.KeyBinding | None:
        """Returns the current key binding, or None if unbound."""

        return self._keybinding

    def get_default_keybinding(self, is_desktop: bool) -> keybindings.KeyBinding | None:
        """Returns the default key binding for the specified layout."""

        return self._desktop_keybinding if is_desktop else self._laptop_keybinding

    def has_default_keybinding(self) -> bool:
        """Returns True if this command has a default keybinding for either layout."""

        return self._desktop_keybinding is not None or self._laptop_keybinding is not None

    def set_keybinding(self, keybinding: keybindings.KeyBinding | None) -> None:
        """Sets the current key binding."""

        self._keybinding = keybinding

    def is_group_toggle(self) -> bool:
        """Returns True if this command toggles its group's enabled state."""

        return self._is_group_toggle

    def is_active(self) -> bool:
        """Returns True if this command should respond to key events."""

        return self._enabled and not self._suspended and self._keybinding is not None


class BrailleCommand(Command):
    """A command that can only be triggered by braille hardware."""

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(
        self,
        name: str,
        function: Callable[..., bool],
        group_label: str,
        description: str = "",
        enabled: bool = True,
        suspended: bool = False,
        braille_bindings: tuple[int, ...] = (),
        executes_in_learn_mode: bool = False,
    ) -> None:
        """Initializes a braille command."""

        super().__init__(name, function, group_label, description, enabled, suspended)
        self._braille_bindings = braille_bindings
        self._executes_in_learn_mode = executes_in_learn_mode

    def __str__(self) -> str:
        """Returns a string representation of the command."""

        parts = [f"BrailleCommand({self._name})"]
        if self._braille_bindings:
            parts.append(f"braille={self._braille_bindings}")
        if self._suspended:
            parts.append("SUSPENDED")
        return " ".join(parts)

    def get_braille_bindings(self) -> tuple[int, ...]:
        """Returns the braille bindings (BrlAPI key codes)."""

        return self._braille_bindings

    def executes_in_learn_mode(self) -> bool:
        """Returns True if this command should execute in learn mode (e.g., pan commands)."""

        return self._executes_in_learn_mode



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for text attribute presentation settings."""

from __future__ import annotations

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, GObject, Gtk

from . import gsettings_registry, guilabels, orca_gui_helpers, text_attribute_manager
from .ax_text import AXTextAttribute
from .ax_utilities import AXUtilities
from .preferences_grid_base import (
    PreferenceActionDoc,
    PreferenceControlDoc,
    PreferencePanelDoc,
    PreferencesGridBase,
    PreferenceValueDoc,
)
from .text_attribute_manager import PresentationMode


class TextAttributePreferencesGrid(PreferencesGridBase):
    """Preferences grid for text attribute presentation settings."""

    # pylint: disable=no-member

    @classmethod
    def get_documentation(cls) -> PreferencePanelDoc:
        """Return documentation metadata for text attribute preferences."""

        return PreferencePanelDoc(
            title=guilabels.TEXT_ATTRIBUTES,
            panel_id="manual.text-attributes",
            description=(
                "Text attributes are font, style, alignment, and other formatting "
                "details associated with text."
                "\n\n"
                + guilabels.TEXT_ATTRIBUTES_INFO
                + "\n\n"
                + "The list contains one row for each text attribute. Use the presentation "
                + "mode on a row to choose whether Orca speaks the attribute, marks it in "
                + "braille, does both, or does neither. Use the row menu or drag handle to "
                + "change the order in which attributes are presented."
            ),
            schema="text-attributes",
            show_available_settings=False,
            controls=(
                PreferenceControlDoc(
                    label=guilabels.TEXT_ATTRIBUTES,
                    kind="table",
                    summary=(
                        "Lists text attributes and lets you choose whether each one is spoken, "
                        "marked in braille, both, or neither."
                    ),
                    value_docs=(
                        PreferenceValueDoc(
                            label=guilabels.TEXT_ATTRIBUTES_PRESENTATION_NONE,
                            summary="Does not present the attribute.",
                        ),
                        PreferenceValueDoc(
                            label=guilabels.PRESENTATION_SPEAK,
                            summary="Speaks the attribute.",
                        ),
                        PreferenceValueDoc(
                            label=guilabels.PRESENTATION_MARK_IN_BRAILLE,
                            summary="Marks the attribute in braille.",
                        ),
                        PreferenceValueDoc(
                            label=guilabels.PRESENTATION_SPEAK_AND_MARK,
                            summary="Speaks the attribute and marks it in braille.",
                        ),
                    ),
                    item_actions=(
                        PreferenceActionDoc(
                            label=guilabels.TEXT_ATTRIBUTES_REORDER,
                            summary=(
                                "Moves the selected text attribute earlier or later in the "
                                "presentation order."
                            ),
                        ),
                        PreferenceActionDoc(
                            label=guilabels.TEXT_ATTRIBUTES_MOVE_TO_TOP,
                            summary="Moves the selected text attribute to the top of the list.",
                        ),
                        PreferenceActionDoc(
                            label=guilabels.TEXT_ATTRIBUTES_MOVE_UP_ONE,
                            summary="Moves the selected text attribute up one position.",
                        ),
                        PreferenceActionDoc(
                            label=guilabels.TEXT_ATTRIBUTES_MOVE_DOWN_ONE,
                            summary="Moves the selected text attribute down one position.",
                        ),
                        PreferenceActionDoc(
                            label=guilabels.TEXT_ATTRIBUTES_MOVE_TO_BOTTOM,
                            summary=(
                                "Moves the selected text attribute to the bottom of the list."
                            ),
                        ),
                    ),
                ),
            ),
        )

    def __init__(self) -> None:
        super().__init__(guilabels.TEXT_ATTRIBUTES)
        self._initializing: bool = True
        self._listbox: Gtk.ListBox | None = None
        self._attributes: list[tuple[AXTextAttribute, PresentationMode]] = []
        self._drag_source_index: int | None = None
        self._focus_target_index: int | None = None

        self._build()
        self._initializing = False
        self.reload()

    def _build(self) -> None:
        """Build the UI components."""

        row = 0

        info_listbox = orca_gui_helpers.create_info_listbox(guilabels.TEXT_ATTRIBUTES_INFO)
        self.attach(info_listbox, 0, row, 1, 1)
        row += 1

        self._listbox = orca_gui_helpers.create_framed_listbox(
            accessible_name=guilabels.TEXT_ATTRIBUTES,
        )
        scrolled_window = orca_gui_helpers.create_preferences_scrolled_window(self._listbox)
        self.attach(scrolled_window, 0, row, 1, 1)
        self.show_all()

    def _create_presentation_mode_model(self) -> Gtk.ListStore:
        """Create the model for presentation mode combo boxes."""

        model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_INT)
        model.append([guilabels.TEXT_ATTRIBUTES_PRESENTATION_NONE, PresentationMode.NONE])
        model.append([guilabels.PRESENTATION_SPEAK, PresentationMode.SPEAK])
        model.append([guilabels.PRESENTATION_MARK_IN_BRAILLE, PresentationMode.BRAILLE])
        model.append([guilabels.PRESENTATION_SPEAK_AND_MARK, PresentationMode.SPEAK_AND_BRAILLE])
        return model

    # pylint: disable-next=too-many-locals,too-many-statements
    def _create_attribute_row(
        self,
        attribute: AXTextAttribute,
        presentation_mode: PresentationMode,
        index: int,
        include_top_separator: bool = True,
    ) -> Gtk.ListBoxRow:
        """Create a ListBoxRow for a text attribute."""

        row = Gtk.ListBoxRow()
        row.set_activatable(False)
        row.get_accessible().set_name(attribute.get_localized_name())

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)

        if include_top_separator:
            separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
            vbox.pack_start(separator, False, False, 0)

        hbox = orca_gui_helpers.create_row_box()

        drag_area = Gtk.EventBox()
        drag_area_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)

        drag_icon = Gtk.Image.new_from_icon_name("list-drag-handle-symbolic", Gtk.IconSize.DND)
        drag_icon.set_opacity(0.5)
        drag_icon.set_margin_end(6)
        drag_area_hbox.pack_start(drag_icon, False, False, 0)

        name_label = Gtk.Label(label=attribute.get_localized_name(), xalign=0)
        name_label.set_hexpand(True)
        drag_area_hbox.pack_start(name_label, True, True, 0)

        drag_area.add(drag_area_hbox)
        hbox.pack_start(drag_area, True, True, 0)

        presentation_combo = Gtk.ComboBox()
        presentation_combo.set_model(self._create_presentation_mode_model())
        renderer = Gtk.CellRendererText()
        presentation_combo.pack_start(renderer, True)
        presentation_combo.add_attribute(renderer, "text", 0)
        presentation_combo.set_active(presentation_mode)
        presentation_combo.connect("changed", self._on_presentation_mode_changed, index)
        hbox.pack_start(presentation_combo, False, False, 0)

        menu_button = Gtk.MenuButton()
        menu_button.set_relief(Gtk.ReliefStyle.NONE)
        icon = Gtk.Image.new_from_icon_name("view-more-symbolic", Gtk.IconSize.DND)
        menu_button.set_image(icon)
        menu_button.get_accessible().set_name(guilabels.TEXT_ATTRIBUTES_REORDER)

        menu = Gtk.Menu()
        move_to_top = Gtk.MenuItem(label=guilabels.TEXT_ATTRIBUTES_MOVE_TO_TOP)
        move_to_top.connect("activate", self._on_menu_move, row, "top")
        menu.append(move_to_top)

        move_up = Gtk.MenuItem(label=guilabels.TEXT_ATTRIBUTES_MOVE_UP_ONE)
        move_up.connect("activate", self._on_menu_move, row, "up")
        menu.append(move_up)

        move_down = Gtk.MenuItem(label=guilabels.TEXT_ATTRIBUTES_MOVE_DOWN_ONE)
        move_down.connect("activate", self._on_menu_move, row, "down")
        menu.append(move_down)

        move_to_bottom = Gtk.MenuItem(label=guilabels.TEXT_ATTRIBUTES_MOVE_TO_BOTTOM)
        move_to_bottom.connect("activate", self._on_menu_move, row, "bottom")
        menu.append(move_to_bottom)

        menu.show_all()
        menu_button.set_popup(menu)
        hbox.pack_end(menu_button, False, False, 0)

        vbox.pack_start(hbox, False, False, 0)
        row.add(vbox)

        row.attribute_index = index
        row.presentation_combo = presentation_combo

        target_entry = Gtk.TargetEntry.new("text/plain", Gtk.TargetFlags.SAME_APP, 0)

        drag_area.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            [target_entry],
            Gdk.DragAction.MOVE,
        )
        drag_area.connect("drag-begin", self._on_drag_begin, row)
        drag_area.connect("drag-data-get", self._on_drag_data_get, row)
        drag_area.connect("drag-end", self._on_drag_end, row)

        row.drag_dest_set(Gtk.DestDefaults.ALL, [target_entry], Gdk.DragAction.MOVE)
        row.connect("drag-data-received", self._on_drag_data_received)

        return row

    def _on_presentation_mode_changed(self, combo: Gtk.ComboBox, index: int) -> None:
        """Handle presentation mode combo box changes."""

        if self._initializing:
            return

        tree_iter = combo.get_active_iter()
        if tree_iter:
            model = combo.get_model()
            mode = model[tree_iter][1]
            self._attributes[index] = (self._attributes[index][0], mode)
            self._has_unsaved_changes = True

    def _on_menu_move(self, _menu_item: Gtk.MenuItem, row: Gtk.ListBoxRow, direction: str) -> None:
        """Handle menu item activation for moving attributes."""

        index = row.attribute_index
        self._move_attribute(index, direction)

    def _move_attribute(self, index: int, direction: str) -> None:
        """Move an attribute in the list."""

        new_index = index

        if direction == "top":
            attribute = self._attributes.pop(index)
            self._attributes.insert(0, attribute)
            new_index = 0
        elif direction == "up" and index > 0:
            self._attributes[index], self._attributes[index - 1] = (
                self._attributes[index - 1],
                self._attributes[index],
            )
            new_index = index - 1
        elif direction == "down" and index < len(self._attributes) - 1:
            self._attributes[index], self._attributes[index + 1] = (
                self._attributes[index + 1],
                self._attributes[index],
            )
            new_index = index + 1
        elif direction == "bottom":
            attribute = self._attributes.pop(index)
            self._attributes.append(attribute)
            new_index = len(self._attributes) - 1

        self._focus_target_index = new_index
        self._has_unsaved_changes = True
        self.refresh()

    def _on_drag_begin(
        self,
        _widget: Gtk.EventBox,
        _drag_context: Gdk.DragContext,
        row: Gtk.ListBoxRow,
    ) -> None:
        """Handle drag begin - store source index."""

        self._drag_source_index = row.attribute_index

    def _on_drag_data_get(
        self,
        _widget: Gtk.EventBox,
        _drag_context: Gdk.DragContext,
        data: Gtk.SelectionData,
        _info: int,
        _time: int,
        row: Gtk.ListBoxRow,
    ) -> None:
        """Handle drag data get - send source index."""

        data.set_text(str(row.attribute_index), -1)

    def _on_drag_data_received(
        self,
        row: Gtk.ListBoxRow,
        _drag_context: Gdk.DragContext,
        _x: int,
        _y: int,
        data: Gtk.SelectionData,
        _info: int,
        _time: int,
    ) -> None:
        """Handle drag data received - perform the move."""

        source_index_str = data.get_text()
        if source_index_str is None:
            return

        try:
            source_index = int(source_index_str)
        except ValueError:
            return

        target_index = row.attribute_index

        if source_index == target_index:
            return

        attribute = self._attributes.pop(source_index)
        self._attributes.insert(target_index, attribute)
        self._has_unsaved_changes = True
        self.refresh()

    def _on_drag_end(
        self,
        _widget: Gtk.EventBox,
        _drag_context: Gdk.DragContext,
        _row: Gtk.ListBoxRow,
    ) -> None:
        """Handle drag end - clean up."""

        self._drag_source_index = None

    def reload(self) -> None:
        """Reload settings from the settings manager and update UI."""

        self._initializing = True
        self._has_unsaved_changes = False

        spoken_attrs = text_attribute_manager.get_manager().get_attributes_to_speak()
        if not spoken_attrs:
            spoken_attrs = [
                attr.get_attribute_name()
                for attr in AXUtilities.get_all_supported_text_attributes()
                if attr.should_present_by_default()
            ]

        brailled_attrs = text_attribute_manager.get_manager().get_attributes_to_braille()
        if not brailled_attrs:
            brailled_attrs = []

        self._attributes = []
        spoken_set = set(spoken_attrs)
        brailled_set = set(brailled_attrs)

        for attr_name in spoken_attrs:
            attr = AXTextAttribute.from_string(attr_name)
            if attr is None:
                continue

            if attr_name in spoken_set and attr_name in brailled_set:
                mode = PresentationMode.SPEAK_AND_BRAILLE
            elif attr_name in spoken_set:
                mode = PresentationMode.SPEAK
            elif attr_name in brailled_set:
                mode = PresentationMode.BRAILLE
            else:
                mode = PresentationMode.NONE

            self._attributes.append((attr, mode))

        for supported_attr in AXUtilities.get_all_supported_text_attributes():
            attr_name = supported_attr.get_attribute_name()
            if attr_name not in spoken_set:
                if attr_name in brailled_set:
                    mode = PresentationMode.BRAILLE
                else:
                    mode = PresentationMode.NONE
                self._attributes.append((supported_attr, mode))

        self._initializing = False
        self.refresh()

    def refresh(self) -> None:
        """Update UI widgets from current state."""

        if self._listbox is None:
            return

        self._initializing = True

        for child in self._listbox.get_children():
            self._listbox.remove(child)

        for index, (attribute, mode) in enumerate(self._attributes):
            row = self._create_attribute_row(
                attribute,
                mode,
                index,
                include_top_separator=index > 0,
            )
            self._listbox.add(row)

        self._listbox.show_all()
        self._initializing = False

        if self._focus_target_index is not None:
            target_row = self._listbox.get_row_at_index(self._focus_target_index)
            if target_row:
                target_row.grab_focus()
            self._focus_target_index = None

    def save_settings(self, profile: str = "", app_name: str = "") -> dict[str, list[str]]:
        """Save current settings and return dict of changed preferences."""

        spoken_attributes = []
        brailled_attributes = []

        for attribute, mode in self._attributes:
            key = attribute.get_attribute_name()
            if mode in (PresentationMode.SPEAK, PresentationMode.SPEAK_AND_BRAILLE):
                spoken_attributes.append(key)
            if mode in (PresentationMode.BRAILLE, PresentationMode.SPEAK_AND_BRAILLE):
                brailled_attributes.append(key)

        result = {
            text_attribute_manager.TextAttributeManager.KEY_ATTRIBUTES_TO_SPEAK: spoken_attributes,
            text_attribute_manager.TextAttributeManager.KEY_ATTRIBUTES_TO_BRAILLE: (
                brailled_attributes
            ),
        }

        if profile and self._has_unsaved_changes:
            skip = not app_name and profile == "default"
            gsettings_registry.get_registry().save_schema(
                "text-attributes",
                result,
                profile,
                app_name,
                skip,
            )

        self._has_unsaved_changes = False
        return result

    # pylint: enable=no-member



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2016-2026 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for system information support."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING

from . import guilabels, preferences_grid_base, system_information_presenter

if TYPE_CHECKING:
    from .system_information_presenter import SystemInformationPresenter


class TimeAndDatePreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Time and Date preferences page."""

    _gsettings_schema = "system-information"
    _documentation_summary = (
        "Use these settings to choose how Orca presents the current date and time."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for system-information preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.KB_GROUP_SYSTEM_INFORMATION,
            panel_id="system_information_presenter.system-information",
            description=(
                "System information settings control how Orca presents information such as "
                "the current date and time."
            ),
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.TIME_AND_DATE,
                    kind="group",
                    summary="Controls how Orca presents the current date and time.",
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.GENERAL_DATE_FORMAT,
                            kind="choice",
                            summary=(
                                "Selects the format Orca uses when presenting the current date."
                            ),
                            schema="system-information",
                            # pylint: disable-next=line-too-long
                            key=system_information_presenter.SystemInformationPresenter.KEY_DATE_FORMAT,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.GENERAL_TIME_FORMAT,
                            kind="choice",
                            summary=(
                                "Selects the format Orca uses when presenting the current time."
                            ),
                            schema="system-information",
                            # pylint: disable-next=line-too-long
                            key=system_information_presenter.SystemInformationPresenter.KEY_TIME_FORMAT,
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, presenter: SystemInformationPresenter) -> None:
        """Initialize the preferences grid."""

        # Generate display options (strftime examples) and values (format strings)
        date_options = []
        date_values = []
        for fmt in system_information_presenter.DateFormat:
            example = time.strftime(fmt.value, time.localtime())
            date_options.append(example)
            date_values.append(fmt.value)

        time_options = []
        time_values = []
        for time_fmt in system_information_presenter.TimeFormat:
            example = time.strftime(time_fmt.value, time.localtime())
            time_options.append(example)
            time_values.append(time_fmt.value)

        controls = [
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.GENERAL_DATE_FORMAT,
                options=date_options,
                values=date_values,
                getter=presenter._get_date_format_string,
                setter=presenter._set_date_format_string,
                prefs_key=system_information_presenter.SystemInformationPresenter.KEY_DATE_FORMAT,
                member_of=guilabels.TIME_AND_DATE,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.GENERAL_TIME_FORMAT,
                options=time_options,
                values=time_values,
                getter=presenter._get_time_format_string,
                setter=presenter._set_time_format_string,
                prefs_key=system_information_presenter.SystemInformationPresenter.KEY_TIME_FORMAT,
                member_of=guilabels.TIME_AND_DATE,
            ),
        ]

        super().__init__(guilabels.KB_GROUP_SYSTEM_INFORMATION, controls)



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for mouse presentation support."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import guilabels, mouse_presenter, preferences_grid_base

if TYPE_CHECKING:
    from .mouse_presenter import MousePresenter


class MousePreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Mouse preferences page."""

    _gsettings_schema = "mouse-review"
    _documentation_summary = (
        "Use these settings to control what Orca presents as the pointer moves."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for mouse preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.MOUSE,
            panel_id="mouse_presenter.mouse",
            description=guilabels.MOUSE_WAYLAND_WARNING,
            schema="mouse-review",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.GENERAL_PRESENT_TOOLTIPS,
                    kind="switch",
                    summary=(
                        "Controls whether Orca presents tooltips that appear when the "
                        "pointer hovers over an object."
                    ),
                    schema="mouse-review",
                    key=mouse_presenter.MousePresenter.KEY_PRESENT_TOOLTIPS,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.GENERAL_SPEAK_OBJECT_UNDER_MOUSE,
                    kind="switch",
                    summary=(
                        "Controls whether Orca speaks the object under the mouse pointer "
                        "as the pointer moves."
                    ),
                    schema="mouse-review",
                    key=mouse_presenter.MousePresenter.KEY_ENABLED,
                ),
            ),
        )

    def __init__(self, presenter: MousePresenter) -> None:
        """Initialize the preferences grid."""

        controls = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.GENERAL_PRESENT_TOOLTIPS,
                getter=presenter.get_present_tooltips,
                setter=presenter.set_present_tooltips,
                prefs_key=mouse_presenter.MousePresenter.KEY_PRESENT_TOOLTIPS,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.GENERAL_SPEAK_OBJECT_UNDER_MOUSE,
                getter=presenter.get_is_enabled,
                setter=presenter.set_is_enabled,
                prefs_key=mouse_presenter.MousePresenter.KEY_ENABLED,
            ),
        ]

        super().__init__(guilabels.MOUSE, controls, info_message=guilabels.MOUSE_WAYLAND_WARNING)



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-lines

"""Preferences grids for braille presentation support."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from . import braille_presenter, gsettings_registry, guilabels, preferences_grid_base

if TYPE_CHECKING:
    from collections.abc import Callable

    from .braille_presenter import BraillePresenter


class BrailleVerbosityPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Braille Verbosity preferences page."""

    _documentation_summary = (
        "Use these settings to choose how much context and object information appears in braille."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for braille verbosity preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.VERBOSITY,
            panel_id="braille_presenter.verbosity",
            description=(
                "Braille verbosity controls how much contextual and object information "
                "Orca shows on the braille display."
            ),
            schema="braille",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.OBJECT_PRESENTATION_IS_DETAILED,
                    kind="switch",
                    summary="Controls whether Orca shows more detailed object information.",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_SHOW_CONTEXT,
                    kind="switch",
                    summary=(
                        "Controls whether Orca shows contextual information about the "
                        "current object, such as the panel it is inside."
                    ),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_DISPLAY_ANCESTORS,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_ABBREVIATED_ROLE_NAMES,
                    kind="switch",
                    summary=(
                        "Controls whether Orca uses abbreviated role names, such as btn for button."
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.PRESENT_OBJECT_MNEMONICS,
                    kind="switch",
                    summary="Controls whether Orca shows object mnemonics in braille.",
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_PRESENT_MNEMONICS,
                ),
            ),
        )

    def __init__(self, presenter: BraillePresenter) -> None:
        self._presenter = presenter
        controls = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.OBJECT_PRESENTATION_IS_DETAILED,
                getter=presenter._get_verbosity_is_detailed,
                setter=presenter._set_verbosity_is_detailed,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.BRAILLE_SHOW_CONTEXT,
                getter=presenter.get_display_ancestors,
                setter=presenter.set_display_ancestors,
                prefs_key=braille_presenter.BraillePresenter.KEY_DISPLAY_ANCESTORS,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.BRAILLE_ABBREVIATED_ROLE_NAMES,
                getter=presenter._get_use_abbreviated_rolenames,
                setter=presenter._set_use_abbreviated_rolenames,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.PRESENT_OBJECT_MNEMONICS,
                getter=presenter.get_present_mnemonics,
                setter=presenter.set_present_mnemonics,
                prefs_key=braille_presenter.BraillePresenter.KEY_PRESENT_MNEMONICS,
            ),
        ]

        super().__init__(guilabels.VERBOSITY, controls)

    def save_settings(self, profile: str = "", app_name: str = "") -> dict[str, Any]:
        """Save settings, writing enum values for verbosity and rolename style."""

        result = super().save_settings(profile, app_name)
        result[braille_presenter.BraillePresenter.KEY_VERBOSITY_LEVEL] = (
            self._presenter.get_verbosity_level()
        )
        rolename_style = self._presenter.get_rolename_style().upper()
        result[braille_presenter.BraillePresenter.KEY_ROLENAME_STYLE] = (
            braille_presenter.VerbosityLevel[rolename_style].value
        )
        return result


class BrailleDisplaySettingsPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Braille Display Settings preferences page."""

    _documentation_summary = (
        "Use these settings to control braille wrapping, contracted braille, and indicator dots."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for braille display settings."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.BRAILLE_DISPLAY_SETTINGS,
            panel_id="braille_presenter.display-settings",
            description=(
                "Display settings control how text is arranged on the braille display, "
                "which braille rules are used, and which dots mark links, selections, "
                "and text attributes."
            ),
            schema="braille",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_ENABLE_END_OF_LINE_SYMBOL,
                    kind="switch",
                    summary=(
                        "Controls whether Orca shows the end-of-line indicator at the "
                        "end of each line of text."
                    ),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_END_OF_LINE_INDICATOR,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_ENABLE_WORD_WRAP,
                    kind="switch",
                    summary=(
                        "Controls whether Orca adjusts displayed text so only full words are "
                        "shown. When word wrap is off, Orca uses all available cells so more "
                        "text can be shown at once."
                    ),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_WORD_WRAP,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_ENABLE_CONTRACTED_BRAILLE,
                    kind="switch",
                    summary=(
                        "Controls whether Orca uses contracted braille, such as shorter "
                        "forms of common letter combinations and words."
                    ),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_CONTRACTED_BRAILLE,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_COMPUTER_BRAILLE_AT_CURSOR,
                    kind="switch",
                    summary=(
                        "When contracted braille is enabled, controls whether the word at "
                        "the cursor is shown in uncontracted computer braille."
                    ),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_COMPUTER_BRAILLE_AT_CURSOR,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_CONTRACTION_TABLE,
                    kind="combo",
                    summary=("Selects the table Orca uses for braille symbols and contractions."),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_CONTRACTION_TABLE,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_INDICATORS,
                    kind="group",
                    summary=(
                        "Controls which braille dots Orca uses to mark links, selections, "
                        "and text attributes. These indicators are shown like underlining "
                        "on the braille display."
                    ),
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.BRAILLE_HYPERLINK_INDICATOR,
                            kind="combo",
                            summary=(
                                "Selects the dot or dots Orca uses to mark text that is "
                                "part of a link."
                            ),
                            schema="braille",
                            key=braille_presenter.BraillePresenter.KEY_LINK_INDICATOR,
                            value_docs=(
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_NONE,
                                    value="none",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_7,
                                    value="dot7",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_8,
                                    value="dot8",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_7_8,
                                    value="dots78",
                                ),
                            ),
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.BRAILLE_SELECTION_INDICATOR,
                            kind="combo",
                            summary=("Selects the dot or dots Orca uses to mark selected text."),
                            schema="braille",
                            key=braille_presenter.BraillePresenter.KEY_SELECTOR_INDICATOR,
                            value_docs=(
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_NONE,
                                    value="none",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_7,
                                    value="dot7",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_8,
                                    value="dot8",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_7_8,
                                    value="dots78",
                                ),
                            ),
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.BRAILLE_TEXT_ATTRIBUTES_INDICATOR,
                            kind="combo",
                            summary=(
                                "Selects the dot or dots Orca uses to mark text attributes, "
                                "such as bold."
                            ),
                            schema="braille",
                            key=braille_presenter.BraillePresenter.KEY_TEXT_ATTRIBUTES_INDICATOR,
                            value_docs=(
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_NONE,
                                    value="none",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_7,
                                    value="dot7",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_8,
                                    value="dot8",
                                ),
                                preferences_grid_base.PreferenceValueDoc(
                                    label=guilabels.BRAILLE_DOT_7_8,
                                    value="dots78",
                                ),
                            ),
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, presenter: BraillePresenter) -> None:
        table_dict = presenter.get_contraction_tables_dict()
        table_names = sorted(table_dict.keys()) if table_dict else []
        table_paths = [table_dict[name] for name in table_names] if table_dict else []

        self._enable_contracted_control = preferences_grid_base.BooleanPreferenceControl(
            label=guilabels.BRAILLE_ENABLE_CONTRACTED_BRAILLE,
            getter=presenter.get_contracted_braille_is_enabled,
            setter=presenter.set_contracted_braille_is_enabled,
            prefs_key=braille_presenter.BraillePresenter.KEY_CONTRACTED_BRAILLE,
        )

        controls: list[
            preferences_grid_base.BooleanPreferenceControl
            | preferences_grid_base.EnumPreferenceControl
        ] = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.BRAILLE_ENABLE_END_OF_LINE_SYMBOL,
                getter=presenter.get_end_of_line_indicator_is_enabled,
                setter=presenter.set_end_of_line_indicator_is_enabled,
                prefs_key=braille_presenter.BraillePresenter.KEY_END_OF_LINE_INDICATOR,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.BRAILLE_ENABLE_WORD_WRAP,
                getter=presenter.get_word_wrap_is_enabled,
                setter=presenter.set_word_wrap_is_enabled,
                prefs_key=braille_presenter.BraillePresenter.KEY_WORD_WRAP,
            ),
            self._enable_contracted_control,
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.BRAILLE_COMPUTER_BRAILLE_AT_CURSOR,
                getter=presenter.get_computer_braille_at_cursor_is_enabled,
                setter=presenter.set_computer_braille_at_cursor_is_enabled,
                prefs_key=braille_presenter.BraillePresenter.KEY_COMPUTER_BRAILLE_AT_CURSOR,
                determine_sensitivity=self._contracted_enabled,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.BRAILLE_CONTRACTION_TABLE,
                options=table_names,
                values=table_paths,
                getter=presenter.get_contraction_table_path,
                setter=presenter.set_contraction_table_from_path,
                prefs_key=braille_presenter.BraillePresenter.KEY_CONTRACTION_TABLE,
                determine_sensitivity=self._contracted_enabled,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.BRAILLE_HYPERLINK_INDICATOR,
                options=[
                    guilabels.BRAILLE_DOT_NONE,
                    guilabels.BRAILLE_DOT_7,
                    guilabels.BRAILLE_DOT_8,
                    guilabels.BRAILLE_DOT_7_8,
                ],
                values=[
                    braille_presenter.BrailleIndicator.NONE.value,
                    braille_presenter.BrailleIndicator.DOT7.value,
                    braille_presenter.BrailleIndicator.DOT8.value,
                    braille_presenter.BrailleIndicator.DOTS78.value,
                ],
                getter=presenter._get_link_indicator_as_int,
                setter=presenter.set_link_indicator_from_int,
                prefs_key=braille_presenter.BraillePresenter.KEY_LINK_INDICATOR,
                member_of=guilabels.BRAILLE_INDICATORS,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.BRAILLE_SELECTION_INDICATOR,
                options=[
                    guilabels.BRAILLE_DOT_NONE,
                    guilabels.BRAILLE_DOT_7,
                    guilabels.BRAILLE_DOT_8,
                    guilabels.BRAILLE_DOT_7_8,
                ],
                values=[
                    braille_presenter.BrailleIndicator.NONE.value,
                    braille_presenter.BrailleIndicator.DOT7.value,
                    braille_presenter.BrailleIndicator.DOT8.value,
                    braille_presenter.BrailleIndicator.DOTS78.value,
                ],
                getter=presenter._get_selector_indicator_as_int,
                setter=presenter.set_selector_indicator_from_int,
                prefs_key=braille_presenter.BraillePresenter.KEY_SELECTOR_INDICATOR,
                member_of=guilabels.BRAILLE_INDICATORS,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.BRAILLE_TEXT_ATTRIBUTES_INDICATOR,
                options=[
                    guilabels.BRAILLE_DOT_NONE,
                    guilabels.BRAILLE_DOT_7,
                    guilabels.BRAILLE_DOT_8,
                    guilabels.BRAILLE_DOT_7_8,
                ],
                values=[
                    braille_presenter.BrailleIndicator.NONE.value,
                    braille_presenter.BrailleIndicator.DOT7.value,
                    braille_presenter.BrailleIndicator.DOT8.value,
                    braille_presenter.BrailleIndicator.DOTS78.value,
                ],
                getter=presenter._get_text_attributes_indicator_as_int,
                setter=presenter.set_text_attributes_indicator_from_int,
                prefs_key=braille_presenter.BraillePresenter.KEY_TEXT_ATTRIBUTES_INDICATOR,
                member_of=guilabels.BRAILLE_INDICATORS,
            ),
        ]

        super().__init__(guilabels.BRAILLE_DISPLAY_SETTINGS, controls)

    def _contracted_enabled(self) -> bool:
        """Check if contracted braille is enabled in the UI."""

        return self._get_control_value(self._enable_contracted_control, True)


class BrailleFlashMessagesPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Braille Flash Messages preferences page."""

    _documentation_summary = "Use these settings to control temporary braille announcements."

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for braille flash message preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.BRAILLE_FLASH_MESSAGES,
            panel_id="braille_presenter.flash-messages",
            description=(
                "Braille flash messages are announcements, such as Orca messages or "
                "command feedback, shown on the braille display. Temporary flash "
                "messages are shown briefly, after which the original display contents "
                "are restored."
            ),
            schema="braille",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_ENABLE_FLASH_MESSAGES,
                    kind="switch",
                    summary="Controls whether Orca shows messages and command feedback in braille.",
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_FLASH_MESSAGES,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_MESSAGES_ARE_DETAILED,
                    kind="switch",
                    summary=(
                        "Controls whether Orca shows detailed braille messages when a "
                        "brief alternative is available. For instance, detailed command "
                        "feedback might include both the setting name and its new value."
                    ),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_FLASH_MESSAGES_DETAILED,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_MESSAGES_ARE_PERSISTENT,
                    kind="switch",
                    summary=(
                        "Controls whether flash messages remain displayed until the display "
                        "is updated instead of disappearing after the configured duration."
                    ),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_FLASH_MESSAGES_PERSISTENT,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_DURATION_SECS,
                    kind="spin",
                    summary=(
                        "Sets how long temporary braille messages are displayed when "
                        "they are not persistent. This setting is ignored when messages "
                        "are persistent."
                    ),
                    minimum="1",
                    maximum="100",
                ),
            ),
        )

    def __init__(self, presenter: BraillePresenter) -> None:
        self._presenter = presenter
        self._flash_persistent_control = preferences_grid_base.BooleanPreferenceControl(
            label=guilabels.BRAILLE_MESSAGES_ARE_PERSISTENT,
            getter=presenter.get_flash_messages_are_persistent,
            setter=presenter.set_flash_messages_are_persistent,
            prefs_key=braille_presenter.BraillePresenter.KEY_FLASH_MESSAGES_PERSISTENT,
        )

        controls: list[
            preferences_grid_base.BooleanPreferenceControl
            | preferences_grid_base.IntRangePreferenceControl
        ] = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.BRAILLE_ENABLE_FLASH_MESSAGES,
                getter=presenter.get_flash_messages_are_enabled,
                setter=presenter.set_flash_messages_are_enabled,
                prefs_key=braille_presenter.BraillePresenter.KEY_FLASH_MESSAGES,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.BRAILLE_MESSAGES_ARE_DETAILED,
                getter=presenter.get_flash_messages_are_detailed,
                setter=presenter.set_flash_messages_are_detailed,
                prefs_key=braille_presenter.BraillePresenter.KEY_FLASH_MESSAGES_DETAILED,
            ),
            self._flash_persistent_control,
            preferences_grid_base.IntRangePreferenceControl(
                label=guilabels.BRAILLE_DURATION_SECS,
                minimum=1,
                maximum=100,
                getter=presenter._get_flash_duration_seconds,
                setter=presenter._set_flash_duration_seconds,
                determine_sensitivity=self._flash_not_persistent,
            ),
        ]

        super().__init__(guilabels.BRAILLE_FLASH_MESSAGES, controls)

    def _flash_not_persistent(self) -> bool:
        """Check if flash messages are not persistent in the UI."""

        return not self._get_control_value(self._flash_persistent_control, False)

    def save_settings(self, profile: str = "", app_name: str = "") -> dict:
        """Persist staged values, including flash-message-duration in milliseconds."""

        result = super().save_settings(profile, app_name)
        result[braille_presenter.BraillePresenter.KEY_FLASH_MESSAGE_DURATION] = (
            self._presenter.get_flash_message_duration()
        )
        return result


class BrailleProgressBarsPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Braille Progress Bars preferences page."""

    _documentation_summary = (
        "Use these settings to decide whether progress-bar changes are shown in braille and "
        "how often they are updated."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for braille progress-bar preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.PROGRESS_BARS,
            panel_id="braille_presenter.progress-bars",
            description=(
                "Progress bar settings control whether Orca shows progress-bar changes "
                "in braille, how often updates are shown, and which progress bars apply."
            ),
            schema="braille",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.GENERAL_BRAILLE_UPDATES,
                    kind="switch",
                    summary="Controls whether Orca periodically shows progress-bar updates.",
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_BRAILLE_PROGRESS_BAR_UPDATES,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.GENERAL_FREQUENCY_SECS,
                    kind="spin",
                    summary="Sets how often Orca shows progress-bar updates.",
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_PROGRESS_BAR_BRAILLE_INTERVAL,
                    minimum="0",
                    maximum="100",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.GENERAL_APPLIES_TO,
                    kind="combo",
                    summary="Selects which progress bars produce braille updates.",
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_PROGRESS_BAR_BRAILLE_VERBOSITY,
                    value_docs=(
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PROGRESS_BAR_ALL,
                            summary="Progress bars in all applications.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PROGRESS_BAR_APPLICATION,
                            summary="Progress bars in the active application.",
                        ),
                        preferences_grid_base.PreferenceValueDoc(
                            label=guilabels.PROGRESS_BAR_WINDOW,
                            summary="Progress bars in the active window.",
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, presenter: BraillePresenter) -> None:
        controls: list[preferences_grid_base.ControlType] = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.GENERAL_BRAILLE_UPDATES,
                getter=presenter.get_braille_progress_bar_updates,
                setter=presenter.set_braille_progress_bar_updates,
                prefs_key=braille_presenter.BraillePresenter.KEY_BRAILLE_PROGRESS_BAR_UPDATES,
            ),
            preferences_grid_base.IntRangePreferenceControl(
                label=guilabels.GENERAL_FREQUENCY_SECS,
                getter=presenter.get_progress_bar_braille_interval,
                setter=presenter.set_progress_bar_braille_interval,
                prefs_key=braille_presenter.BraillePresenter.KEY_PROGRESS_BAR_BRAILLE_INTERVAL,
                minimum=0,
                maximum=100,
            ),
            preferences_grid_base.EnumPreferenceControl(
                label=guilabels.GENERAL_APPLIES_TO,
                getter=presenter.get_progress_bar_braille_verbosity,
                setter=presenter.set_progress_bar_braille_verbosity,
                prefs_key=braille_presenter.BraillePresenter.KEY_PROGRESS_BAR_BRAILLE_VERBOSITY,
                options=[
                    guilabels.PROGRESS_BAR_ALL,
                    guilabels.PROGRESS_BAR_APPLICATION,
                    guilabels.PROGRESS_BAR_WINDOW,
                ],
                values=[
                    braille_presenter.ProgressBarVerbosity.ALL.value,
                    braille_presenter.ProgressBarVerbosity.APPLICATION.value,
                    braille_presenter.ProgressBarVerbosity.WINDOW.value,
                ],
            ),
        ]

        super().__init__(guilabels.PROGRESS_BARS, controls)


class BrailleOSDPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the braille on-screen display preferences page."""

    _documentation_summary = "Use these settings to control the visual braille monitor window."

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for braille on-screen display preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.ON_SCREEN_DISPLAY,
            panel_id="braille_presenter.on-screen-display",
            description=guilabels.BRAILLE_MONITOR_INFO,
            schema="braille",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_MONITOR_CELL_COUNT,
                    kind="spin",
                    summary=(
                        "Sets the number of braille cells shown in the on-screen braille display."
                    ),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_MONITOR_CELL_COUNT,
                    minimum="1",
                    maximum="80",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_MONITOR_SHOW_DOTS,
                    kind="switch",
                    summary=(
                        "Controls whether the on-screen braille display shows Unicode "
                        "braille dot patterns instead of text characters."
                    ),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_MONITOR_SHOW_DOTS,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_MONITOR_FOREGROUND,
                    kind="color",
                    summary="Selects the text color for the on-screen braille display.",
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_MONITOR_FOREGROUND,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_MONITOR_BACKGROUND,
                    kind="color",
                    summary=("Selects the background color for the on-screen braille display."),
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_MONITOR_BACKGROUND,
                ),
            ),
        )

    def __init__(self, presenter: BraillePresenter) -> None:
        controls: list[preferences_grid_base.ControlType] = [
            preferences_grid_base.IntRangePreferenceControl(
                label=guilabels.BRAILLE_MONITOR_CELL_COUNT,
                getter=presenter.get_monitor_cell_count,
                setter=presenter.set_monitor_cell_count,
                prefs_key=braille_presenter.BraillePresenter.KEY_MONITOR_CELL_COUNT,
                minimum=1,
                maximum=80,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.BRAILLE_MONITOR_SHOW_DOTS,
                getter=presenter.get_monitor_show_dots,
                setter=presenter.set_monitor_show_dots,
                prefs_key=braille_presenter.BraillePresenter.KEY_MONITOR_SHOW_DOTS,
            ),
            preferences_grid_base.ColorPreferenceControl(
                label=guilabels.BRAILLE_MONITOR_FOREGROUND,
                getter=presenter.get_monitor_foreground,
                setter=presenter.set_monitor_foreground,
                prefs_key=braille_presenter.BraillePresenter.KEY_MONITOR_FOREGROUND,
            ),
            preferences_grid_base.ColorPreferenceControl(
                label=guilabels.BRAILLE_MONITOR_BACKGROUND,
                getter=presenter.get_monitor_background,
                setter=presenter.set_monitor_background,
                prefs_key=braille_presenter.BraillePresenter.KEY_MONITOR_BACKGROUND,
            ),
        ]

        super().__init__(
            guilabels.ON_SCREEN_DISPLAY,
            controls,
            info_message=guilabels.BRAILLE_MONITOR_INFO,
        )


# pylint: disable-next=too-many-instance-attributes
class BraillePreferencesGrid(preferences_grid_base.PreferencesGridBase):
    """GtkGrid containing the Braille preferences page with nested stack navigation."""

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for the braille preferences landing page."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.BRAILLE,
            panel_id="manual.braille",
            description=(
                "Braille settings control what Orca presents in braille. The list "
                "contains braille-related settings and pages.\n\n"
                "Activate a row, or press Right Arrow, to open its settings. Press "
                "Left Arrow, Escape, or Alt+Left to return to the list."
            ),
            show_available_settings=False,
            schema="braille",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_ENABLE_BRAILLE_SUPPORT,
                    kind="switch",
                    summary="Turns braille output on or off.",
                    schema="braille",
                    key=braille_presenter.BraillePresenter.KEY_ENABLED,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.VERBOSITY,
                    kind="page",
                    summary="Opens settings for how much information Orca shows in braille.",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_DISPLAY_SETTINGS,
                    kind="page",
                    summary="Opens settings for braille wrapping, contraction, and indicators.",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.BRAILLE_FLASH_MESSAGES,
                    kind="page",
                    summary="Opens settings for temporary braille messages.",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.PROGRESS_BARS,
                    kind="page",
                    summary="Opens settings for when Orca shows progress-bar updates in braille.",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.ON_SCREEN_DISPLAY,
                    kind="page",
                    summary="Opens settings for the on-screen braille display.",
                ),
            ),
        )

    def __init__(
        self,
        presenter: BraillePresenter,
        title_change_callback: Callable[[str], None] | None = None,
    ) -> None:
        super().__init__(guilabels.BRAILLE)
        self._presenter = presenter
        self._initializing = True
        self._title_change_callback = title_change_callback

        self._verbosity_grid = BrailleVerbosityPreferencesGrid(presenter)
        self._display_settings_grid = BrailleDisplaySettingsPreferencesGrid(presenter)
        self._flash_messages_grid = BrailleFlashMessagesPreferencesGrid(presenter)
        self._progress_bars_grid = BrailleProgressBarsPreferencesGrid(presenter)
        self._osd_grid = BrailleOSDPreferencesGrid(presenter)

        self._build()
        self._initializing = False

    def _build(self) -> None:
        """Build the nested stack UI."""

        row = 0

        categories = [
            (guilabels.VERBOSITY, "verbosity", self._verbosity_grid),
            (guilabels.BRAILLE_DISPLAY_SETTINGS, "display_settings", self._display_settings_grid),
            (guilabels.BRAILLE_FLASH_MESSAGES, "flash_messages", self._flash_messages_grid),
            (guilabels.PROGRESS_BARS, "progress-bars", self._progress_bars_grid),
            (guilabels.ON_SCREEN_DISPLAY, "osd", self._osd_grid),
        ]

        enable_listbox, stack, _categories_listbox = self._create_child_grid_preferences_stack(
            enable_label=guilabels.BRAILLE_ENABLE_BRAILLE_SUPPORT,
            enable_getter=self._presenter.get_braille_is_enabled,
            enable_setter=self._presenter.set_braille_is_enabled,
            categories=categories,
            title_change_callback=self._title_change_callback,
            main_title=guilabels.BRAILLE,
        )

        self.attach(enable_listbox, 0, row, 1, 1)
        row += 1
        self.attach(stack, 0, row, 1, 1)

    def on_becoming_visible(self) -> None:
        """Reset to the categories view when this grid becomes visible."""

        self.multipage_on_becoming_visible()

    def reload(self) -> None:
        """Fetch fresh values and update UI."""

        self._initializing = True
        self._has_unsaved_changes = False
        self._verbosity_grid.reload()
        self._display_settings_grid.reload()
        self._flash_messages_grid.reload()
        self._progress_bars_grid.reload()
        self._osd_grid.reload()
        self._initializing = False

    def save_settings(self, profile: str = "", app_name: str = "") -> dict:
        """Persist staged values."""

        assert self._multipage_enable_switch is not None
        result: dict[str, Any] = {}
        result[braille_presenter.BraillePresenter.KEY_ENABLED] = (
            self._multipage_enable_switch.get_active()
        )
        result.update(self._verbosity_grid.save_settings())
        result.update(self._display_settings_grid.save_settings())
        result.update(self._flash_messages_grid.save_settings())
        result.update(self._progress_bars_grid.save_settings())
        result.update(self._osd_grid.save_settings())

        if profile:
            skip = not app_name and profile == "default"
            gsettings_registry.get_registry().save_schema(
                "braille",
                result,
                profile,
                app_name,
                skip,
            )

        return result

    def refresh(self) -> None:
        """Update widgets from staged values."""

        self._initializing = True
        self._verbosity_grid.refresh()
        self._display_settings_grid.refresh()
        self._flash_messages_grid.refresh()
        self._progress_bars_grid.refresh()
        self._osd_grid.refresh()
        self._initializing = False

    def has_changes(self) -> bool:
        """Return True if there are unsaved changes."""

        return (
            self._has_unsaved_changes
            or self._verbosity_grid.has_changes()
            or self._display_settings_grid.has_changes()
            or self._flash_messages_grid.has_changes()
            or self._progress_bars_grid.has_changes()
            or self._osd_grid.has_changes()
        )
